"""
For ussd we need two store two things.
    1. users ussd session
    2. customer journey.


For ussd session each session has its own record and all users use the same journey.


users ussd session store
========================
comming soon


Customer journey store
======================

.. automodule:: ussd.store.journey_store

"""
