__author__ = 'Mwaaas'
__email__ = 'francismwangi152@gmail.com'
__phone_number__ = '+254702729654'


from marshmallow import fields

fields.Field.default_error_messages["required"] = 'This field is required.'
